import GPminer as GPm
import pandas as pd
import numpy as np
import FreeBack as FB


# 因子命名格式：
# factor-method-period
# factor: 基础因子名， 如果有参数，用.分割
# method：时序算符, period: 时序算符周期
class Factor():
    def __init__(self, market):
        self.market = market
        for i in ['close', 'open', 'high', 'low', 'pre_close', 'amount', 'vol']:
            if i not in market.columns:
                raise ValueError("value must include %s"%i)
        if 'exfactor' not in market.columns:
            self.cal_factor('exfactor')
    # 计算/引用因子
    def cal_factor(self, code):
        if code in self.market.columns:
            return self.market[code]
        else:
            exp = code.split('-')
            if len(exp)==1:  # 基础因子
                self.cal_basic_factor(code)
            else:  # 时序计算（MA/EMA/Zscore/Max/Min/Std/Sum/rank/quantile/Skew/Kurt/argmin/argmax/prod)
                self.market[code] = FB.my_pd.cal_ts(self.ref(exp[0]), exp[1], int(exp[2])) 
                return self.market[code]
    # 计算/引用基础因子
    def cal_basic_factor(self, code):
        if code in self.market.columns:
             return         # 如果因子存在则不计算
        code_split = code.split('.')
        key = code_split[0]
        para = code_split[1:]
        if key=='exfactor':    # 复权因子
            exfactor = self.cal_factor('close').groupby('code').shift()/self.cal_factor('pre_close')
            self.market[code] = exfactor.groupby('code').cumprod()
        # 量价基础 basic
        elif key in ['ex_close', 'ex_open', 'ex_high', 'ex_low']:
            self.market[code] = self.cal_factor(code[3:])*self.cal_factor('exfactor')
        elif key=='vwap':     
            self.market[code] = self.cal_factor('amount')/self.cal_factor('vol')
        # 动量反转  mom
        elif key=='Ret':      
            self.market[code] = self.cal_factor('close')/self.cal_factor('pre_close')
        elif key=='nightRet':     # 隔夜收益（开盘跳空
            self.market[code] = self.cal_factor('open')/self.cal_factor('pre_close')-1
        elif key=='T0Ret':
            self.market['T0Ret'] = self.cal_factor('close')/self.cal_factor('open')-1
        elif key=='maxRet':
            self.market['maxRet'] = self.cal_factor('high')/self.cal_factor('low')-1
        elif key=='bodyRatio':   # k线实体占比
            self.market[code] = abs(self.cal_factor('close')-self.cal_factor('open'))/(self.cal_factor('high')-self.cal_factor('low'))
        elif key=='RSI':     
            up = pd.Series(np.where(self.cal_factor('Ret')>0, self.cal_factor('Ret'), 0), \
                           index=self.market.index)
            down = pd.Series(np.where(self.cal_factor('Ret')<0, -self.cal_factor('Ret'), 0), \
                             index=self.market.index)
            sumup = FB.my_pd.cal_ts(up, 'Sum', int(para[0]))
            sumdown = FB.my_pd.cal_ts(down, 'Sum', int(para[0]))
            rsi = sumup/(sumup+sumdown)
            self.market[code] = rsi
        elif key=='DrawDown':        # 唐奇安通道 上轨距离   DrawDown.d
            max_c = FB.my_pd.cal_ts(self.cal_factor('ex_high'), 'Max', int(para[0]))
            self.market[code] = self.cal_factor('ex_close')/min_c-1 
        elif key=='DrawUp':   # 唐奇安通道 下轨距离  DrawUp.d
            min_c = FB.my_pd.cal_ts(self.cal_factor('ex_low'), 'Min', int(para[0]))
            self.market[code] = 1-self.cal_factor('ex_close')/max_c
        # 波动率  volatility
        elif key=='AMP':   # 振幅   AMP.d
            exhigh_Max = FB.my_pd.cal_ts(self.cal_factor('ex_high'), 'Max', int(para[0]))
            exlow_Min = FB.my_pd.cal_ts(self.cal_factor('ex_low'), 'Min', int(para[0]))
            self.market[code] = (exhigh_Max - exlow_Min)/exlow_Min
        elif key=='TR':    # 真实波动
            self.market[code] = pd.concat([self.cal_factor('high')-self.cal_factor('low'),\
                abs(self.cal_factor('pre_close')-self.cal_factor('high')), \
                abs(self.cal_factor('pre_close')-self.cal_factor('low'))], axis=1).max(axis=1) 
        # 成交流动性   vol
        elif key=='Amihud':  # 流动性
            self.market[code] = self.cal_factor('Ret')/self.cal_factor('amount')
        elif key=='UnusualVol':    # 异常成交量  UnusualVol.d
            self.market[code] = self.cal_factor('vol')/self.cal_factor('vol_MA_'+para[0])
        # 技术形态  kbars
        elif key=='MACD':   #  MACD.d.d.d 
            EMAslow = FB.my_pd.cal_ts(self.cal_factor('ex_close'), 'EMA', int(para[1])) 
            EMAfast = FB.my_pd.cal_ts(self.cal_factor('ex_close'), 'EMA', int(para[0]))
            DIF = EMAfast-EMAslow
            DEM = FB.my_pd.cal_ts(DIF, 'EMA', int(para[2]))
            self.market[code] = DIF-DEM
        elif key=='RSRS':        # high/low 回归斜率  RSRS.d
            deltax = self.cal_factor('ex_low')-FB.my_pd.cal_ts(self.cal_factor('ex_low'), 'MA', int(para[0]))
            deltay = self.cal_factor('ex_high')-FB.my_pd.cal_ts(self.cal_factor('ex_high'), 'MA', int(para[0]))
            self.market[code] = FB.my_pd.cal_ts(deltax*deltay, 'Sum', int(para[0]))/FB.my_pd.cal_ts(deltax**2, 'Sum', int(para[0])) 
        # 量价相关  composite
        elif key in ['VolVWAPCorr', 'VolAMPCorr', 'VolRetCorr']:  # 量价相关性   VolPriceCorr.d
            deltax = self.cal_factor('vol')-FB.my_pd.cal_ts(self.cal_factor('vol'), 'MA', int(para[0]))
            if key=='VolVWAPCorr':            
                deltay = self.cal_factor('vwap')-FB.my_pd.cal_ts(self.cal_factor('vwap'), 'MA', int(para[0]))
            elif key=='VolAMPCorr':
                deltay = self.cal_factor('AMP.1')-FB.my_pd.cal_ts(self.cal_factor('AMP.1'), 'MA', int(para[0]))
            elif key=='VolRetCorr':
                deltay = self.cal_factor('Ret')-FB.my_pd.cal_ts(self.cal_factor('Ret'), 'MA', int(para[0]))
            self.market[code] = (FB.my_pd.cal_ts((deltax*deltay), 'Sum', int(para[0]))/\
                    (np.sqrt(FB.my_pd.cal_ts(deltax**2, 'Sum', int(para[0])))*\
                     np.sqrt(FB.my_pd.cal_ts(deltay**2, 'Sum', int(para[0]))))).fillna(0)
        # 金融学理论   model        
        elif key=='beta':    # CAMP理论 beta/alpha/estd  beta.d
            # 计算市场平均收益
            deltax = self.cal_factor('meanRet')-FB.my_pd.cal_ts(self.cal_factor('meanRet'), 'MA', int(para[0]))
            deltay = self.cal_factor('Ret')-FB.my_pd.cal_ts(self.cal_factor('Ret'), 'MA', int(para[0]))
            self.market[code] = FB.my_pd.cal_ts(deltax*deltay, 'Sum', int(para[0]))/FB.my_pd.cal_ts(deltax**2, 'Sum', int(para[0])) 
        elif key=='oc2hl':   # 开收重心到低高重心变动
            self.market[code] = (self.cal_factor('high')+self.cal_factor('low'))/(self.cal_factor('open')+self.cal_factor('close')) 
        return self.market[code]



import numpy as np
import pandas as pd
from random import shuffle,choice,sample
import GPminer as GPm
from itertools import combinations
import time, copy

# 种群繁殖类

class Gen():
    # 因子库（针对score和pool可以单独指定因子库），种群，市场，种群的类型
    def __init__(self, basket=[], popu0=None, market=None, indtype=GPm.ind.Score,\
                        score_basket=None, pool_basket=None):
        if (score_basket==None)&(pool_basket==None):
            self.score_basket = self.pool_basket = self.basket = basket
        elif score_basket==None:
            self.score_basket = self.basket = basket
            self.pool_basket = pool_basket
        elif pool_basket==None:
            self.pool_basket = self.basket = basket
            self.score_basket = score_basket
        else:
            self.score_basket = score_basket
            self.pool_basket = pool_basket
        if popu0==None:
            self.popu = GPm.popu.Population(indtype)
        else:
            self.popu = popu0
        # 初始化pool的参数域，需要输入market
        if  (self.popu.type==GPm.ind.Pooland) | (self.popu.type==GPm.ind.Pool) |\
            (self.popu.type==GPm.ind.SP):
            if type(market)==type(None):
                print('market is needed for Pool ind Gen!')
                return 
            self.para_space = {}
            for factor in self.pool_basket:
                # 数值因子，小于等于divide_n个数时全部因子值进入参数空间
                divide_n = 100
                if type(market[factor].iloc[0]) in [np.float64, np.int64, type(1.0), type(1)]:
                    if len(market[factor].unique())>divide_n:
                        self.para_space[factor] = (False, [market[factor].quantile(i) \
                                    for i in np.linspace(0.01,0.99,divide_n)]) 
                    else:  # 最大最小值去除
                        self.para_space[factor] = (False, sorted(market[factor].unique())[1:-1]) 
                else:
                    self.para_space[factor] = (True, list(market[factor].unique())) 
    # 从basket中因子获得popu
    def get_seeds(self, exclude=True):
        def seeds_Score():
            popu0 = GPm.popu.Population() 
            # 遍历单因子、双因子, 作为种子组合
            for i in self.score_basket:
                popu0.add({GPm.ind.Score([[i, True, 1]]).code, GPm.ind.Score([[i, False, 1]]).code})
            for i,j in list(combinations(self.score_basket, 2)):
                for b0 in [True, False]:
                    for b1 in [True, False]:
                        popu0.add(GPm.ind.Score([[i, b0, 1], [j, b1, 1]]).code)
            return popu0
        # 仅生成排除因子
        def seeds_Pool():
            popu0 = GPm.popu.Population(GPm.ind.Pool)
            # 单因子组合
            for factor in self.pool_basket:
                for threshold in self.para_space[factor][1]:
                    # 离散变量使用=
                    if self.para_space[factor][0]:
                        if exclude:
                            pool0 = GPm.ind.Pool([[], [['equal', factor, [threshold, ]]]])
                        else:
                            pool0 = GPm.ind.Pool([[['equal', factor, [threshold, ]]], []])
                        popu0.add(pool0.code)
                    else:
                        if exclude:
                            pool0 = GPm.ind.Pool([[], [['less', factor, threshold]]])
                        else:
                            pool0 = GPm.ind.Pool([[['less', factor, threshold]], []])
                        popu0.add(pool0.code)
                        if exclude:
                            pool0 = GPm.ind.Pool([[], [['greater', factor, threshold]]])
                        else:
                            pool0 = GPm.ind.Pool([[['greater', factor, threshold]], []])
                        popu0.add(pool0.code)
            popu0 = popu0.subset(int(0.1*len(popu0.codes)))
            # 两因子组合
            if exclude:
                combos = [GPm.ind.Pool(i[0] + '|' + i[1][1:]).code \
                                 for i in combinations(popu0.codes, 2)]
            else:
                combos = [GPm.ind.Pool(i[0][:-1] + '|' + i[1]).code \
                                 for i in combinations(popu0.codes, 2)]
            popu0.add(set(sample(combos, int(0.1*len(combos)))))
            popu0.add(set(combos))
            return popu0
        if self.popu.type==(GPm.ind.Score):
            return seeds_Score().codes
        # Pool和Pooland的code/exp是互通的
        elif  (self.popu.type==GPm.ind.Pooland) | (self.popu.type==(GPm.ind.Pool)):
            return seeds_Pool().codes
        elif self.popu.type==(GPm.ind.SP):
            return set(i[0]+'&'+i[1] for i in zip(seeds_Score().codes, seeds_Pool().codes))
    # 增大或减小某因子参数
    def mutation_d(self, ind):
        exp = copy.deepcopy(ind.exp)
        if type(ind)==GPm.ind.Score:
            # 单因子权重无法改变
            if len(exp)==1:
                return ind
            random_select = np.random.randint(len(exp)) 
            GPm.ino.log('选择变异%s第%s个因子权重'%(ind.code, random_select))
            deltawmax = 0.1 # 权重改变幅度小于此阈值   # 声明常数 constant
            deltawmin = 0.02 # 权重改变幅度大于此阈值
            max_step = 100 # 寻找新权重组合最大尝试次数
            # 全部权重乘mul，随机选一个因子权重+-d(\
            # 改变random_select的权重可以通过 1.改变该因子的乘数 或 2.改变其他因子的乘数实现)
            sumw = sum([i[2] for i in exp])
            wbefore = exp[random_select][2]/sumw  # 因子权重
            mul = d = 1  
            # 使用mehtod、opt、mul、d改变乘数后该因子权重
            def get_wafter(method=True, opt=False):
                if method:
                    if opt:
                        for i in range(len(exp)):
                            exp[i][2] = exp[i][2]*mul-d
                        exp[random_select][2] = exp[random_select][2]+d
                    return (mul*exp[random_select][2])/(mul*sumw-(len(exp)-1)*d)
                else:
                    if opt:
                        for i in range(len(exp)):
                            exp[i][2] = exp[i][2]*mul
                        exp[random_select][2] = exp[random_select][2]+d
                    return (mul*exp[random_select][2]+d)/(mul*sumw+d)
            minw = min([i[2] for i in exp])
            # 增大、减小权重50%概率
            if np.random.rand()>0.5:
                d = 1
                # 如果最小数字*mul大于d则通过减小乘数增大权重，否则选择通过增大系数增大权重
                wafter = get_wafter(minw*mul>d, False)
                step = 0
                while ((wafter-wbefore<deltawmin)|(wafter-wbefore>deltawmax))&(step<max_step): 
                    if (wafter-wbefore)<deltawmax:
                        # 权重变化太小增大d
                        d+=1
                    else:
                        # 权重变化太大增大mul
                        mul+=1
                    step += 1
                    # GPm.ino.log(mul, d)
                    wafter = get_wafter(minw*mul>d, False)
                get_wafter(minw*mul>d, True)
            else:
                d = -1
                # 如果最小数字*mul小于等于-d的话选择通过增大系数减小权重(d<0)
                wafter = get_wafter(minw*mul<=-d, False)
                step = 0
                while ((wbefore-wafter<deltawmin)|(wbefore-wafter>deltawmax))&(step<max_step): 
                    if (wbefore-wafter)<deltawmax:
                        # 权重变化太小减小d
                        d-=1
                    else:
                        # 权重变化太大增大mul
                        mul+=1
                        step += 1
                    wafter = get_wafter(minw*mul<=-d, False)
                get_wafter(minw*mul<=-d, True)
                #GPm.ino.log('通过%s系数, 减小权重, mul=%s, d=%s'%\
                #      ((lambda x: '减小' if x else '增大')(method), mul, d))
            new = GPm.ind.Score(exp)
            return new
        elif  (type(ind)==GPm.ind.Pooland) | (type(ind)==GPm.ind.Pool):
            # 等概率选择变异include或exclude部分（除非他们为空）
            if (len(exp[0])!=0)&((np.random.rand()<0.5)|(len(exp[1])==0)):
                select_inexlude = 0
                select_loc = np.random.randint(len(exp[0]))
            else:
                select_inexlude = 1
                select_loc = np.random.randint(len(exp[1]))
            thisfactor_space = self.para_space[exp[select_inexlude][select_loc][1]]
            # 对于离散型因子增加或减少value
            if thisfactor_space[0]:
                if len(exp[select_inexlude][select_loc][2])==1:
                    exp[select_inexlude][select_loc][2].append(choice(thisfactor_space[1]))
                else:
                    if np.random.rand()<0.5:
                        exp[select_inexlude][select_loc][2].append(choice(thisfactor_space[1]))
                    else:
                        exp[select_inexlude][select_loc][2].pop()
            # 对于数值型因子改变value到临近值
            else:
                less_value = sorted([i for i in thisfactor_space[1] if i<exp[select_inexlude][select_loc][2]])
                larger_value = sorted([i for i in thisfactor_space[1] if i>exp[select_inexlude][select_loc][2]])
                if (less_value == []) & (larger_value == []):
                    print('没有可变异的值')
                elif less_value==[]:
                    exp[select_inexlude][select_loc][2] = larger_value[0]
                elif larger_value==[]:
                    exp[select_inexlude][select_loc][2] = less_value[-1]
                else:
                    if np.random.rand()<0.5:
                        exp[select_inexlude][select_loc][2] = larger_value[0]
                    else:
                        exp[select_inexlude][select_loc][2] = less_value[-1]
            new = type(ind)(exp)
            return new
        elif type(ind)==GPm.ind.SP:
            # 随机选打分因子/排除因子变异
            if np.random.rand()<0.5:
                score = self.mutation_d(ind.score)
                return GPm.ind.SP(score.code+'&'+ind.pool.code)
            else:
                pool = self.mutation_d(ind.pool)
                return GPm.ind.SP(ind.score.code+'&'+pool.code)
        return 'pass all' 
    def popu_mutation_d(self):
        # 随机取出一个个体，变异得到新个体，添加得到个体。
        ind = self.popu.subset()
        ind = self.mutation_d(ind)
        self.popu.add(ind.code)
    # 增减因子
    def mutation_and(self, ind):
        exp = copy.deepcopy(ind.exp)
        if type(ind)==GPm.ind.Score:
            random_select0 = np.random.randint(len(exp))
            if (np.random.rand()>0.5)&(len(exp)!=1):
                exp.pop(random_select0)
            else:
                random_select = np.random.randint(len(self.score_basket))
                # 随机赋一个已有因子的权重
                exp.append([self.score_basket[random_select], np.random.rand()>0.5, \
                             exp[random_select0][2]])
            new = GPm.ind.Score(exp)
            return new
        elif (type(ind)==GPm.ind.Pooland) | (type(ind)==GPm.ind.Pool):
            def expand(exp):
                exp = copy.deepcopy(exp)
                if (np.random.rand()>0.5)&(len(exp)!=1):
                    exp.pop()
                else:
                    random_factor = choice(self.pool_basket)
                    if self.para_space[random_factor][0]:
                        exp.append(['equal', random_factor, [choice(self.para_space[random_factor][1])]])
                    else:
                        exp.append(['less' if np.random.rand()>0.5 else 'greater', random_factor, \
                             choice(self.para_space[random_factor][1])])
                return exp
            if exp[0]==[]:
                exp = [[], expand(exp[1])]
            elif exp[1]==[]:
                exp = [expand(exp[0]), []]
            else:
                if np.random.rand()<0.5:
                    exp = [expand(exp[0]), exp[1]]
                else:
                    exp = [exp[0], expand(exp[1])]
            new = self.popu.type(exp)
            return new
        elif type(ind)==GPm.ind.SP:
            # 随机选打分因子/排除因子变异
            if np.random.rand()<0.5:
                score = self.mutation_and(ind.score)
                return GPm.ind.SP(score.code+'&'+ind.pool.code)
            else:
                pool = self.mutation_and(ind.pool)
                return GPm.ind.SP(ind.score.code+'&'+pool.code)
    def popu_mutation_and(self):
        ind = self.popu.subset()
        ind = self.mutation_and(ind)
        self.popu.add(ind.code)
    # 替换因子
    def mutation_replace(self, ind):
        exp = copy.deepcopy(ind.exp)
        if type(ind)==GPm.ind.Score:
            random_select0 = np.random.randint(len(exp))
            random_select = np.random.randint(len(self.score_basket))
            already = [i[0] for i in exp]
            # 不能替换前后相同，不能替换已有因子
            while (exp[random_select0][0]==self.score_basket[random_select])|\
                            (self.score_basket[random_select] in already):
                random_select = np.random.randint(len(self.score_basket))
            exp[random_select0][0] = self.score_basket[random_select]
            return GPm.ind.Score(exp)
        elif (type(ind)==GPm.ind.Pooland) | (type(ind)==GPm.ind.Pool):
            # 删一个加一个
            def expreplace(exp):
                exp.pop()
                random_factor = choice(self.pool_basket)
                # 随机赋一个para_space中阈值
                if self.para_space[random_factor][0]: 
                    exp.append(['equal', random_factor, \
                                    [choice(self.para_space[random_factor][1])]])
                else:
                    exp.append(['less' if np.random.rand()>0.5 else 'greater',\
                                 random_factor, choice(self.para_space[random_factor][1])])
                return exp
            if exp[0]==[]:
                exp = [[], expreplace(exp[1])]
            elif exp[1]==[]:
                exp = [expreplace(exp[0]), []]
            else:
                if np.random.rand()<0.5:
                    exp = [[], expreplace(exp[1])]
                else:
                    exp = [expreplace(exp[1]), []]
            return type(ind)(exp)
        else:
            # 随机选打分因子/排除因子变异
            if np.random.rand()<0.5:
                score = self.mutation_replace(ind.score)
                return GPm.ind.SP(score.code+'&'+ind.pool.code)
            else:
                pool = self.mutation_replace(ind.pool)
                return GPm.ind.SP(ind.score.code+'&'+pool.code)
    def popu_mutation_replace(self):
        ind = self.popu.subset()
        ind = self.mutation_replace(ind)
        self.popu.add(ind.code)
    # 合成因子
    def mutation_sum(self, ind0, ind1):
        pass
    # 种群繁殖
    def multiply(self, multi=2, prob_dict={}):
        # 各算子被执行的概率，如果空则全部算子等概率执形
        if prob_dict=={}:
            opts = [f for f in dir(Gen) if  ('popu' in f)]
            prob_ser = pd.Series(np.ones(len(opts)), index=opts)
        else:
            prob_ser = pd.Series(prob_dict.values(), index=prob_dict.keys())
        prob_ser = prob_ser/prob_ser.sum()
        prob_ser = prob_ser.cumsum()
        # 种群繁殖到目标数量，同时限制单次变异最大时间5s
        popu_size = len(self.popu.codes)
        from func_timeout import func_set_timeout
        @func_set_timeout(5)
        def run_mul(func):
            getattr(self, func)()
        while len(self.popu.codes)<int(popu_size*multi):
            r = np.random.rand()
            #GPm.ino.log('算子选择随机数：%.3lf'%r)
            func = prob_ser[prob_ser>r].index[0]
            try:
                run_mul(func)
                #GPm.ino.log('执行完毕')
            except:
                GPm.ino.log('warning!!! %s超过最大运行时间5s'%func)





import math
import numpy as np 
import pandas as pd
from random import shuffle,sample

# 种群的个体单元
# 可以通过code或exp生成，code和exp一一对应，code用于hash，exp用于计算。
class Ind():
    def __init__(self, input=""):
        if input=="":
            self.exp=[]
            self.code=""
        elif type(input)==type(""):
            self.code = input
            self.code2exp()
            self.uexp()
            self.exp2code()
        elif type(input)==type([]):
            self.exp = input
            self.uexp()
            self.exp2code()
        else:
            self.exp = None
            self.code = None
    # 从code获得exp
    def code2exp(self):
        pass
    # 从exp获得code
    def exp2code(self):
        pass
    # 保证表达式唯一性
    def uexp(self):
        pass

# 打分因子, 
# code:'2*False*a+1*True*b, exp:[[2, False, 'a'], [1, True, 'b']]
# 2倍a因子越小越好+1倍b因子越大越好
class Score(Ind):
    max_exp_len = 10 # 最大因子数
    max_mul = 50 # 最大因子系数          constant
    rankall = False # 池子外股票是否参与排序
    def code2exp(self):
        exp = []
        split = self.code.split('+')
        for i in split:
            # 组合中的单个因子
            one = i.split('*')
            one = [one[2], one[1]=='True', int(one[0])]
            exp.append(one)
        self.exp = exp
    def exp2code(self):
        self.code = '+'.join([str(int(i[2]))+'*'+str(i[1])+'*'+i[0] for i in self.exp])
        self.score = self
    # 保证等价的表达式唯一
    def uexp(self):
        exp = [] 
        already = []
        shuffle(self.exp)  # 当因子重复出现时添加一些随机性
        for i in self.exp:
            # 如果出现0或负值则直接跳过
            if i[2]<=0:
                continue
            # 如果同一因子出现两次以上则只保留第一个
            elif i[0] not in already:
                already.append(i[0])
                exp.append(i)
        exp = exp[:self.max_exp_len]
        # 除以最大公因数
        ws = [i[2] for i in exp] 
        smaller=int(min(ws))
        for i in reversed(range(1,smaller+1)):
            if list(filter(lambda j: j%i!=0,ws)) == []:
                hcf = i
                break
        # 限制最大同时考虑等权情况
        exp = [[i[0], i[1], min(i[2]/hcf, self.max_mul)] for i in exp]
        if max([i[2] for i in exp])==min([i[2] for i in exp]):
            exp = [[i[0], i[1], 1] for i in exp]
        # 先按权重排序，再按因子名称排序，再按方向排序
        def takewsort(one):
            return one[::-1]
        exp.sort(key=takewsort, reverse=True)
        self.exp = exp
    # 获取其包含的全部因子
    def factors(self):
        count = {}
        for i in self.exp:
            try:
                count[i[0]] += i[2]
            except:
                count[i[0]] = i[2]
        return pd.Series(count).sort_values(ascending=False)
    # 比较打分因子大小
    def compare(self, s0):
        # 先比较因子数量
        ns = s0.count('+')
        # 再比较权重数字大小
        ws = sum([float(i.split('*')[0]) for i in s0.split('+')]) 
        return 1e5*ns+ws 
    def short(self):
        return ','.join([i[0] for i in self.exp])

# 排除/选取因子（确定策略池子）
# code：'a<130|b=A,B|c>C'  exp:[[['less', 'a', 130]], [['equal', 'b', ['A', 'B']], ['great', 'c', 'C']] 
# ;前为include，后为exclude条件, 意为全部a<130的股票中排除掉b为A和B以及c大于C的股票。
class Pool(Ind):
    max_exp_len = 10 # 最大因子数
    def code2exp(self):
        innex = self.code.split(';')
        final_exp = []
        for code in innex:
            exp = []
            if code=='':
                final_exp.append(exp)
                continue
            split = code.split('|')
            for i in split:
                # 组合中的单个条件
                if '<' in i:
                    opt = 'less'
                    s = i.split('<')
                    # 可排序非数值数据
                    try:
                        value = float(s[1])
                    except:
                        value = s[1]
                    factor = s[0]
                elif '>' in i:
                    opt='greater'
                    s = i.split('>')
                    try:
                        value = float(s[1])
                    except:
                        value = s[1]
                    factor = s[0]
                else:
                    opt='equal'
                    s = i.split('=')
                    value = []
                    for i in s[1].split(','):
                        try:
                            value.append(float(i))
                        except:
                            value.append(i)
                    factor = s[0]
                one = [opt, factor, value]
                exp.append(one)
            final_exp.append(exp)
        self.exp = final_exp
    def exp2code(self):
        code = []
        for exp in self.exp:
            code.append('|'.join([i[1]+\
                (lambda x:'<' if x=='less' else '>' if x=='greater' else '=')(i[0]) +\
                 (str(i[2]) if type(i[2])!=list else ','.join([str(j) for j in i[2]])) for i in exp]))
        self.code = ';'.join(code)
        self.pool = self
    def uexp(self):
        def unique_c(exp):
            # 先按因子名称排序，再按逻辑符号，再按值
            def takewsort(one):
                return [one[1], one[0], str(one[2])]
            exp.sort(key=takewsort, reverse=True)
            # 大小于号重叠部分去除，等于重复去除
            prefactor = ''
            preopt = ''
            prevalue = 0
            unique_exp = []
            for c in exp:
                opt = c[0]
                factor = c[1]
                value = c[2]
                #print('因子和操作符都相同时需要考虑合并问题')
                if (factor==prefactor)&(opt==preopt):
                    if opt=='equal':   # 新元素全部并入
                        unique_exp.pop()
                        unique_exp.append([opt, factor, sorted([str(i) for i in list(values|set(value))])])
                        values = values|set(value)
                        continue
                    elif opt=='less':
                        # 如果集合扩大则更新
                        if value>max(values):
                            unique_exp.pop()
                            unique_exp.append(c)
                    elif opt=='greater':
                        if value<min(values):
                            unique_exp.pop()
                            unique_exp.append(c)
                    values.append(value)
                else:
                    #print('因子相同，操作符不同时，可能表达式代表的是全集，如a<10|a>5则随机去掉一个条件)
                    if (factor==prefactor)&(opt!=preopt)&(opt!='equal'):
                        if (((opt=='less')&(value>=prevalue))|((opt=='greater')&(value<=prevalue))):
                            if np.random.rand()<0.5:
                                pass  # 去掉该条件
                            else:
                                unique_exp.pop()   # 去掉前一个条件
                                unique_exp.append(c)
                            continue
                    preopt = opt
                    prefactor = factor
                    prevalue = value
                    if opt=='equal':
                        if type(value)==list:
                            values = set(value)
                        else:
                            values = set([value])
                        unique_exp.append([opt, factor, sorted([str(v) for v in values])])
                        #unique_exp.append([opt, factor, sorted(values)])
                    else:
                        values = [value, ]
                        unique_exp.append(c)
            # 限制公式最大长度
            if len(unique_exp)>self.max_exp_len:
                unique_exp = [unique_exp[i] for i in sorted(sample(range(len(unique_exp)), \
                                                               self.max_exp_len))]
            return unique_exp
        self.exp = [unique_c(self.exp[0]), unique_c(self.exp[1])]
    def factors(self):
        count = {}
        for i in self.exp:
            for j in i:
                try:
                    count[j[1]] += 1 
                except:
                    count[j[1]] = 1 
        return pd.Series(count).sort_values(ascending=False)
    def around(self, para_space):
        def a(exp_):
            new = []
            for e in exp_:
                # 离散变量，扔掉不在para_space中值
                if para_space[e[1]][0]:
                    new.append([e[0], e[1], [i for i in e[2] if i in para_space[e[1]][1]]])
                # 连续变量，如果值不在para_space中的话，取最接近的
                else:
                    if e[2] in para_space[e[1]][1]:
                        new.append([e[0], e[1], e[2]])
                        continue 
                    try:
                        big = [v for v in para_space[e[1]][1] if v>e[2]][0]
                    except:  # 边界值
                        new.append([e[0], e[1], para_space[e[1]][1][-1]])
                        continue
                    try:
                        small = [v for v in para_space[e[1]][1] if v<e[2]][-1]
                    except:
                        new.append([e[0], e[1], para_space[e[1]][1][0]])
                        continue
                    new.append([e[0], e[1], small if (big-e[2])>=(e[2]-small) else big])
            return new
        return Pool([a(self.exp[0]), a(self.exp[1])])
    def short(self):
        return ','.join([i[1] for i in self.exp[0]])+\
                    ';'+','.join([i[1] for i in self.exp[1]])

# 交集池子
class Pooland(Pool):
    def uexp(self):
        def unique_c(exp):
            # 先按因子名称排序，再按逻辑符号，再按值
            def takewsort(one):
                return [one[1], one[0], str(one[2])]
            exp.sort(key=takewsort, reverse=True)
            # 大小于号重叠部分保留，等于重复保留
            prefactor = ''
            preopt = ''
            prevalue = 0
            unique_exp = []
            for c in exp:
                opt = c[0]
                factor = c[1]
                value = c[2]
                #print('因子和操作符都相同时需要考虑合并问题')
                if (factor==prefactor)&(opt==preopt):
                    if opt=='equal':  
                        unique_exp.pop()
                        if values&set(value):
                            unique_exp.append([opt, factor, \
                                            sorted([str(i) for i in list(values&set(value))])])
                        values = values&set(value)
                        continue
                    elif opt=='less':
                        # 如果集合缩小则更新
                        if value>values:
                            continue
                        else:
                            unique_exp.pop()
                            unique_exp.append(c)
                            values = value
                    elif opt=='greater':
                        if value<values:
                            continue
                        else:
                            unique_exp.pop()
                            unique_exp.append(c)
                            values = value
                else:
                    #print('因子相同，操作符不同时，可能表达式代表的是空集，如a<10|a>20则随机去掉一个条件)
                    if (factor==prefactor)&(opt!=preopt)&(opt!='equal'):
                        if (((opt=='less')&(value<=prevalue))|((opt=='greater')&(value>=prevalue))):
                            if np.random.rand()<0.5:
                                pass  # 去掉该条件
                            else:
                                unique_exp.pop()   # 去掉前一个条件
                                unique_exp.append(c)
                            continue
                    preopt = opt
                    prefactor = factor
                    prevalue = value
                    if opt=='equal':
                        values = set(value)
                        unique_exp.append([opt, factor, sorted([str(v) for v in values])])
                    else:
                        values = value
                        unique_exp.append(c)
            # 限制公式最大长度
            if len(unique_exp)>self.max_exp_len:
                unique_exp = [unique_exp[i] for i in sorted(sample(range(len(unique_exp)), \
                                                               self.max_exp_len))]
            return unique_exp
        self.exp = [unique_c(self.exp[0]), unique_c(self.exp[1])]

# 策略类，包含Score和Pool
# code: Score.code+'&'+Pool.code, exp: [Score.exp, Pool.exp]
class SP(Ind):
    def __init__(self, input=None):
        self.score = self.pool = None
        super().__init__(input)
    def code2exp(self):
        if self.score==self.pool==None:        
            scorecode, poolcode = self.code.split('&')
            self.score = Score(scorecode)
            self.pool = Pool(poolcode)
        self.exp = [self.score.exp, self.pool.exp]
    def exp2code(self):
        if self.score==self.pool==None:        
            scoreexp, poolexp = self.exp
            self.score = Score(scoreexp)
            self.pool = Pool(poolexp)
        self.code = self.score.code+'&'+self.pool.code
    def factors(self):
        return self.score.factors().add(self.pool.factors(), fill_value=0).sort_values(ascending=False)
    def short(self):
        return self.score.short()+'&'+self.pool.short()



import pandas as pd
import numpy as np
from random import sample
from joblib import Parallel, delayed
import GPminer as GPm
import FreeBack as FB
import time, datetime, os


class Miner():
    # 矿工初始化需输入策略超参数，share表示挖掘出一组策略的共享部分，
    # 如果share是pool则挖掘score，反之亦然，如果是None则挖掘SP
    # 对于Pool与SP挖掘，可以设置最大排除比例，如果挖掘出的排除因子超出此比例则适应度直接赋为最差
    def __init__(self, market, benchmark=None, share=None, pool_basket=None, score_basket=None, p0=None,\
                  hold_num=5, comm=10/1e4, price='close', code_returns=None, max_extract=1, fixp=None):
        self.market = market
        self.benchmark = benchmark
        self.share = share
        if type(self.share)==GPm.ind.Pool:
            self.indtype = GPm.ind.Score
        elif type(self.share)==GPm.ind.Score:
            self.indtype = GPm.ind.Pool
        else:
            self.indtype = GPm.ind.SP
        if pool_basket!=None:
            self.pool_basket = pool_basket
        else:
            self.pool_basket = list(market.columns)
        if score_basket!=None:
            self.score_basket = score_basket
        else:
            self.score_basket = list(market.columns)
        self.p0 = p0
        self.hold_num = hold_num
        self.comm = comm
        self.price = price
        self.code_returns = code_returns
        self.max_extract = max_extract
        self.fixp = fixp
    def prepare(self, fitness='sharpe',\
                 population_size=10, evolution_ratio=0.2, tolerance_g=3, max_g=10,\
                  prob_dict={}, select_alg='cut', n_core=4, exclude=True):
        self.fitness = fitness
        self.population_size = population_size
        self.evolution_ratio = evolution_ratio
        self.tolerance_g = tolerance_g
        self.max_g = max_g
        self.prob_dict = prob_dict
        self.select_alg = select_alg
        self.n_core = n_core
        # 生成初代种群需要
        popu0 = GPm.popu.Population(type=self.indtype, fix_ind=self.fixp)
        self.gen0 = GPm.gen.Gen(score_basket=self.score_basket, pool_basket=self.pool_basket,\
                            market=self.market, indtype=self.indtype, popu0=popu0)
        if type(self.p0)==type(None):
            self.seeds = list(self.gen0.get_seeds(exclude=exclude))
        else:
            if type(self.p0)==type(set()):
                self.gen0.popu.add(self.p0)
            else:
                self.gen0.popu.add(self.p0.code)
            #while len(self.gen0.popu.codes)<int(self.population_size/self.evolution_ratio):
            while len(self.gen0.popu.codes)<int(10*self.population_size):
                self.gen0.multiply()
            self.seeds = list(self.gen0.popu.codes)
    def run(self, pooltype='or'):
        workfile = datetime.datetime.now().strftime("%m%d%H%M_%S_%f")+\
                            '_%s'%np.random.rand()
        t0 = time.time()
        if type(self.p0)==type(None):
            #init_seeds = sample(self.seeds, int(self.population_size/self.evolution_ratio))
            init_seeds = sample(self.seeds, int(self.population_size))
        elif type(self.p0)==type(set()):
            #init_seeds = set(sample(self.seeds, int(self.population_size/self.evolution_ratio)))
            init_seeds = set(sample(self.seeds, int(self.population_size)))
        else:
            #init_seeds = set(sample(self.seeds, int(self.population_size/self.evolution_ratio)-1))|\
            #        {self.p0.code}
            init_seeds = set(sample(self.seeds, int(self.population_size)-1))|\
                    {self.p0.code}
        GPm.ino.log('生成%s个p作为初始种群'%len(init_seeds))
        GPm.ino.log('=====此初始种群进化开始=====')
        os.mkdir(workfile)
        fitness_all = pd.DataFrame()
        fitness_df = pd.DataFrame()
        # 后续进化在popu0上操作 
        popu0 = GPm.popu.Population(type=self.indtype, fix_ind=self.fixp)
        if self.indtype==GPm.ind.Score:
            eval0 = GPm.eval.Eval(self.market, pool=self.share)
            if pooltype=='or':
                eval0.eval_pool()
            elif pooltype=='and':
                eval0.eval_pool(mod='and')
        elif self.indtype==GPm.ind.Pool:
            eval0 = GPm.eval.Eval(self.market, score=self.share)
        else:
            eval0 = GPm.eval.Eval(self.market)
        for ind in init_seeds:
            popu0.add(ind) 
        gen0 = GPm.gen.Gen(score_basket=self.score_basket, pool_basket=self.pool_basket, market=self.market,\
                            indtype=self.indtype, popu0=popu0)
        # 计算适应度
        def single(p):
            result = pd.DataFrame(columns=['return_total', 'return_annual', 'excess_annual',\
                    'sharpe', 'excess_sharpe', 'drawdown', 'excess_drawdown', \
                    'sigma', 'excess_sigma', 'beta', 'alpha', 'extract_ratio'])
            if self.indtype==GPm.ind.Score:
                eval0.eval_score(p)
            elif (self.indtype==GPm.ind.Pool) | (self.indtype==GPm.ind.Pooland):
                if pooltype=='or':
                    eval0.eval_pool(p)
                elif pooltype=='and':
                    eval0.eval_pool(p, mod='and')
                if eval0.market['include'].mean()<1-self.max_extract:
                    result.loc[p, :] = -99999
                    result.loc[p, 'extract_ratio'] = 1-eval0.market['include'].mean()
                    return result
                eval0.eval_score()
            else:
                psplit = p.split('&')
                if pooltype=='or':
                    eval0.eval_pool(psplit[1])
                elif pooltype=='and':
                    eval0.eval_pool(psplit[1], mod='and')
                if eval0.market['include'].mean()<1-self.max_extract:
                    result.loc[p, :] = -99999
                    return result
                eval0.eval_score(psplit[0])
            strat0 = eval0.backtest(self.hold_num, self.price, self.code_returns)
            post0 = FB.post.StratPost(strat0, eval0.market, benchmark=self.benchmark,\
                                        comm=self.comm, show=False)
            result.loc[p, 'return_total'] = post0.return_total
            result.loc[p, 'return_annual'] = post0.return_annual
            result.loc[p, 'excess_annual'] = post0.excess_return_annual
            result.loc[p, 'sharpe'] = post0.sharpe
            result.loc[p, 'excess_sharpe'] = post0.excess_sharpe
            result.loc[p, 'drawdown'] = -max(post0.drawdown)
            result.loc[p, 'excess_drawdown'] = -max(post0.excess_drawdown)
            result.loc[p, 'sigma'] = -post0.sigma
            result.loc[p, 'excess_sigma'] = -post0.excess_sigma
            result.loc[p, 'beta'] = post0.beta
            result.loc[p, 'alpha'] = post0.alpha*250*100
            result.loc[p, 'extract_ratio'] = 1-eval0.market['include'].mean()
            return result
        max_fitness = -99999
        max_loc = 0
        for g in range(self.max_g):
            GPm.ino.log('第%s代'%(g))
            if g!=0:
                # 之前已经计算过的无需计算
                fitness_df = fitness_all.loc[list(popu0.codes&set(fitness_all.index))]
            GPm.ino.log('本代%d个策略，其中%d个策略已有计算结果'%(len(popu0.codes), len(fitness_df)))
            if len(popu0.codes)!=len(fitness_df):
                if self.n_core!=1:   # 并行
                    fitness_list = Parallel(n_jobs=self.n_core)(delayed(single)(p) \
                                            for p in list(popu0.codes-set(fitness_df.index)))
                else:    # 串行
                    fitness_list = []
                    for p in list(popu0.codes - set(fitness_df.index)):
                        fitness_list.append(single(p))
                fitness_all = pd.concat([fitness_all, pd.concat(fitness_list)]).drop_duplicates()
                fitness_df = pd.concat([fitness_df, pd.concat(fitness_list)])
            GPm.ino.log('第%s轮进化适应度计算完成'%g)
            fitness_df = fitness_df.sort_values(self.fitness, ascending=False)
            if fitness_df.iloc[0][self.fitness]>max_fitness:
                max_fitness = fitness_df.iloc[0][self.fitness]
                max_loc = g
            fitness_df.to_csv(workfile+'/fitness%s.csv'%g)
            # 选择
            if g!=0:
                if self.select_alg=='cut':
                    popu0.reset(set(fitness_df[:self.population_size].index)) # 截断选择
                # 锦标赛，不放回
                elif self.select_alg=='tournament':
                    select = set()
                    while len(select)<self.population_size:
                        one = set(fitness_df.loc[sample(list(set(fitness_df.index)-select), int(len(fitness_df)/10))]\
                                    .sort_values(by=self.fitness, ascending=False).index[:1])
                        select = select|one
                    popu0.reset(select)
            GPm.ino.log('第%s轮进化完成，最大%s:%.2lf'%(g, self.fitness, fitness_df.iloc[0][self.fitness]))
            if ((g-max_loc)>=self.tolerance_g)|(g>=(self.max_g-1)):
                cost = time.time()-t0
                GPm.ino.log('=====此初始种群进化完成=====共计算%d个策略，总耗时%.1lfs，单策略耗时%.2lfs'%(\
                    len(fitness_all), cost, cost/len(fitness_all)))
                fitness_df.loc[list(popu0.codes)].sort_values(by=self.fitness, ascending=False).\
                    to_csv(workfile+'/fitness%s.csv'%(g+1))
                # 重命名结果
                os.rename(workfile, 'result-'+self.indtype(fitness_df.index[0]).short()+'-'+workfile)
                break
            # 种群繁殖
            gen0.multiply(1/self.evolution_ratio)
            GPm.ino.log('交叉变异生成第%s代种群'%(g+1))





import dill,os,pickle,datetime


# 反序列化一个对象

def save_pkl(obj,filename, fileloc, serial=False):
    if serial:
        obj = dill.dumps(obj)
    if type(obj)==type(set()):
        type_name = '.set'
    elif type(obj)==type(list()):
        type_name = '.list'
    elif type(obj)==type(dict()):
        type_name = '.dict'
    with open(fileloc+'/'+filename+type_name+'.pkl','wb') as file:
        pickle.dump(obj,file)
def read_pkl(filename, fileloc, serial=False):
    filename = [i for i in os.listdir(fileloc) if filename in i][0]
    with open(fileloc+'/'+filename,'rb') as file:
        obj = pickle.load(file)
        if serial:
            obj = dill.loads(obj)
        return obj

# log 函数
def log(*txt):
    try:
        #try:
        #    f = open('log.txt','a+', encoding='gbk')
        #except:
        #    f = open('log.txt','a+', encoding='utf-8')
        f = open('log.txt','a+', encoding='utf-8')
        write_str = ('\n'+' '*35).join([str(i) for i in txt])
        f.write('%s,        %s\n' % \
            (datetime.datetime.now(), write_str))
        f.close()
    except PermissionError as e:
        print(f"Error: {e}. You don't have permission to access the specified file.")



import pandas as pd
import numpy as np
import FreeBack as FB
import GPminer as GPm 
import time

# 计算个体适应度类

class Eval():
    def __init__(self, market, pool=None, score=None):
        self.market = market
        self.pool = pool
        self.score = score
    def eval_pool(self, poolcode=None, mod='or'):
        if poolcode!=None:
            if mod=='or':
                self.pool = GPm.ind.Pool(poolcode)
            elif mod=='and':
                self.pool = GPm.ind.Pooland(poolcode)
        # 如果pool为None则include=True
        if self.pool!=None:
            # 默认全包含
            if self.pool.exp[0]!=[]:
                result = []
                for c in self.pool.exp[0]:
                    if c[0]=='less':
                        r=(self.market[c[1]]<c[2])
                    elif c[0]=='greater':
                        r=(self.market[c[1]]>c[2])
                    elif c[0]=='equal':
                        r=(self.market[c[1]].isin(c[2]))
                    result.append(r)
                if mod=='or':
                    include = pd.concat(result, axis=1).any(axis=1)
                elif mod=='and':
                    include = pd.concat(result, axis=1).all(axis=1)
            else:
                include = pd.Series(True, index=self.market.index)
            # 默认不排除
            if self.pool.exp[1]!=[]:
                result = []
                for c in self.pool.exp[1]:
                    if c[0]=='less':
                        r=(self.market[c[1]]<c[2])
                    elif c[0]=='greater':
                        r=(self.market[c[1]]>c[2])
                    elif c[0]=='equal':
                        r=(self.market[c[1]].isin(c[2]))
                    result.append(r)
                if mod=='or':
                    exclude = pd.concat(result, axis=1).any(axis=1)
                elif mod=='and':
                    exclude = pd.concat(result, axis=1).all(axis=1)
            else:
                exclude = pd.Series(False, index=self.market.index)
            self.market['include'] = include&(~exclude)
        else:
            self.market['include'] = True 
    def eval_score(self, scorecode=None):
        #time0 = time.time()
        if scorecode!=None:
            self.score = GPm.ind.Score(scorecode)
        # 获取筛选/排除后factor排序
        def process_factor(factor_name):
            if self.score.rankall:
                return self.market[factor_name]
            else:
                # 排除掉的置为np.nan
                return self.market[factor_name]*(self.market['include'].astype(int).\
                                                replace(to_replace={0:np.nan}))
        # 获取打分
        for factor in self.score.exp:
            self.market[factor[0]+'_score'] = process_factor(factor[0]).groupby('date').\
                rank(ascending=factor[1])*factor[2]
        basescore = [i[0]+'_score' for i in self.score.exp]
        self.market['score'] = self.market[basescore].sum(axis=1)
        #ino.log('获取打分耗时', time.time()-time0)
        #time0 = time.time()
    def backtest(self, hold_num, price, code_returns=None, interval=1):
        # 回测
        strat0 = FB.strat.MetaStrat(self.market, 'include', 'score',\
                    hold_num, price, interval=interval, code_returns=code_returns)
        strat0.run()
        return strat0





import pandas as pd
from random import choice,sample
import GPminer as GPm


# 种群
class Population():
    # 使用一个集合来储存code值, fix是一个个体，为该种群所有个体公用的个体，默认为空
    def __init__(self, type=GPm.ind.Score, fix_ind=None):
        self.type = type
        self.codes = set()
        self.fix_ind = fix_ind
    def code2exp(self, code):
        return self.type(code)
    # 附加上固定属性
    def add_fix(self, code):
        if self.fix_ind==None:
            return code
        elif '&' in code:  # SP
            s, p = code.split('&')
            s_, p_ = self.fix_ind.code.split('&')
            return self.add_fix(code.split('&')[0]) + '&' + self.add_fix(code.split('&')[1])
        elif ';' in code:  # Pool/Pooland
            _ = self.fix_ind.pool.code.split(';')
            __ = code.split(';')
            __[0] = ('|' if (_[0]!='')&(__[0]!='') else '') + __[0]
            __[1] = ('|' if (_[1]!='')&(__[1]!='') else '') + __[1]
            return _[0]+__[0]+';'+_[1]+__[1]
        else:        # Score
            if code=='':
                return self.fix_ind.score.code
            elif self.fix_ind.score.code=='':
                return code
            else:
                return code + '+' + self.fix_ind.score.code
    # 默认单个输入的时候检查ind是否重复，输入集合时不检查
    def add(self, code):
        if type(code)!=type(set()):
            self.codes = self.codes|{self.type(self.add_fix(code)).code}
        else:
            self.codes = self.codes|code
    def sub(self, code):
        if type(code)!=type(set()):
            self.codes = self.codes-{self.type(code).code}
        else:
            self.codes = self.codes-code
    def reset(self, code):
        self.codes = set()
        self.add(code)
    def get_name(self, n=3):
        factor_count = pd.Series()  # 因子出现频率
        for i in self.codes:
            factor_count = factor_count.add(self.type(i).factors(), fill_value=0)
        self.name = ';'.join(factor_count.sort_values(ascending=False).index[:n]) 
        return self.name
    # 从群体中采样
    def subset(self, size=1):
        if size==1:
            return self.type(sample(list(self.codes), 1)[0])
        popu0 = Population(self.type)
        popu0.add(set(sample(list(self.codes), size)))
        return popu0





from setuptools import setup, find_packages
from codecs import open

with open("README.md","r",encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name="GPminer",
    # 版本号: 第几次模块增加，第几次函数增加，第几次函数功能修改
    # (每次高级别序号增加后,低级别序号归0)
    # alpha为调试版,beta为测试版,没有后缀为稳定版 
    version="4.0.0",
    author="LH.Li",
    author_email="lh98lee@zju.edu.cn",  
    description='Package for factor strategy mining by GP algorithm',
    long_description=long_description, 
    # 描述文件为md格式
    long_description_content_type="text/markdown",
    url="https://github.com/LHanLi/GPminer",
    packages=find_packages(),
    install_requires = [
        #'pandas',
        #'scipy',
        #'statsmodels',
        #'seaborn',
        #'plottable',
        #'pyecharts',
        #'numpy_ext',
        #'xlsxwriter'
    ],
    classifiers=[
         # 该软件包仅与Python3兼容
        "Programming Language :: Python :: 3",
        # 根据GPL 3.0许可证开源
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        # 与操作系统无关
        "Operating System :: OS Independent",
    ],
)

# python3 setup.py install --u
